/*
 * Программный комплекс Artix:SCO является интеллектуальной собственностью
 * ИП Заболотная Е.А. и охраняется законом согласно ст. 1225 ГК РФ.
 * Запрещается любое использование программного комплекса Artix:SCO, как целиком, так
 * и его отдельных компонентов (API, интерфейс, сервисы, алгоритмы и др.), без
 * письменного разрешения ИП Заболотная Е.А.
 */

// eslint-disable-next-line no-unused-vars
const debug = (msg, obj) => console.log(msg, JSON.stringify(obj, null, 2))

// Задержка в мс
// Promise не работает как надо https://bugreports.qt.io/browse/QTBUG-71329
function setTimeout(timeout, func) {
    // eslint-disable-next-line no-undef
    const timer = new Timer()
    timer.interval = timeout
    timer.singleShot = true
    timer.timeout.connect(this, function () { func() })
    timer.start()
}

// Вызов api отменен
class Cancelled {
    constructor() {
        this.name = "Cancelled"
    }
}

/*
 * Справочники
 */

// Товар в справочнике
class Product {
    constructor(props) {
        this.name = ""                     // наименование
        this.barcode = ""                  // штрих-код
        this.price = 0                     // цена
        this.unit = ""                     // единица измерения количества
        this.fractional = false            // весовой товар
        this.discount = 0                  // процент скидки
        this.requestQuantity = false       // требуется ввод количества
        this.requestWeight = false         // требуется взвешивание
        this.visualVerify = false          // требуется визуальная проверка
        this.lowWeight = false             // легкий товар
        this.bypassWeightControl = false   // не ожидать товар на контрольных весах
        this.age = 0                       // ограничение продажи по возрасту
        this.egais = false                 // запрос акцизной марки ЕГАИС
        this.mark = false                  // запрос маркировки товара
        this.unity = false                 // объединять данные позиции в чеке
        this.posVerify = false             // необходима проверка на стороне POS
        this.tags = []                     // тэги товара
        Object.assign(this, props)
    }
}

// Создание объекта справочника товаров из списка
function objListToDict(keyName, list) {
    let dict = {}
    for (const i in list) {
        const obj = list[i]
        dict[obj[keyName]] = obj
    }
    return dict
}

// Справочник товаров
const products = objListToDict("barcode", [
    new Product({ barcode: "1", name: "Тестовый товар 1", price: 1000, unit: "шт" }),
    new Product({ barcode: "2", name: "Тестовый товар 2", price: 21900 }),
    new Product({ barcode: "3", name: "Тестовый товар 3", price: 9950 }),
    new Product({ barcode: "4", name: "На рубль с вводом количества", price: 100, requestQuantity: true }),
    new Product({ barcode: "5", name: "На копейку с вводом количества", price: 1, requestQuantity: true }),
    new Product({ barcode: "6", name: "Весовой товар", price: 43912, fractional: true }),
    new Product({ barcode: "7", name: "Переход в оплату с ошибкой", price: 38164 }),
    new Product({ barcode: "8", name: "Возврат в формирование с ошибкой", price: 19273 }),
    new Product({ barcode: "9", name: "Товар со скидкой 30%", price: 9822, discount: 30 }),
    new Product({ barcode: "10", name: "Ошибка оплаты банковской картой", price: 32938 }),
    new Product({ barcode: "11", name: "Ошибка закрытия чека", price: 2763 }),
    new Product({ barcode: "12", name: "Ошибка при удалении", price: 42347 }),
    new Product({ barcode: "13", name: "Ошибка при отмене чека", price: 9827 }),
    new Product({ barcode: "14", name: "Нельзя менять количество", price: 9827 }),
    new Product({ barcode: "15", name: "Требуется взвешивание", price: 2200, fractional: true, requestWeight: true, tags: ["scophotocontrol"] }),
    new Product({ barcode: "16", name: "Легкий товар", price: 2300, lowWeight: true }),
    new Product({ barcode: "17", name: "Пропускается при контроле веса", price: 2400, bypassWeightControl: true }),
    new Product({ barcode: "18", name: "Подтверждение возраста", price: 46524, age: 18 }),
    new Product({ barcode: "19", name: "Визуальный контроль", price: 90937, visualVerify: true }),
    new Product({ barcode: "20", name: "Алкоголь", price: 39037, age: 18, egais: true }),
    new Product({ barcode: "21", name: "Запрос маркировки товара", price: 9737, mark: true }),
    new Product({ barcode: "22", name: "Объединяемая позиция", price: 8654, unity: true }),
    new Product({ barcode: "23", name: "Требуется проверка позиции на стороне клиента", price: 30641, posVerify: true }),
    new Product({ barcode: "24", name: "Отправка чека на Email", price: 1922 }),
    new Product({ barcode: "25", name: "Ошибка оплаты банковской картой для кассира", price: 37938 }),
    new Product({ barcode: "26", name: "Товар с печатью купона", price: 12312 }),
    new Product({ barcode: "392", name: "АИ-92", price: 2275, unit: "л", fractional: true, requestWeight: true, bypassWeightControl: true }),
    new Product({ barcode: "395", name: "АИ-92", price: 2450, unit: "л", fractional: true, requestWeight: true, bypassWeightControl: true }),
    new Product({ barcode: "305", name: "ДТ", price: 2520, unit: "л", fractional: true, requestWeight: true, bypassWeightControl: true }),
    new Product({ barcode: "123", name: "Пакет", price: 453 }),
])

// Пользователь
class User {
    constructor(props) {
        this.id = ""     // id
        this.name = ""   // имя
        Object.assign(this, props)
    }
}

// Справочник пользователей
const users = objListToDict("id", [
    new User({ id: "1", name: "Кассир" }),
    new User({ id: "2", name: "Старший кассир" }),
    new User({ id: "3", name: "Мастер" }),
])
let user = null

// Подарочная карта
class GiftCard {
    constructor(props) {
        this.barcode = ""    // штрих-код
        this.sum = ""        // сумма
        Object.assign(this, props)
    }
}

// Справочник подарочных карт
const giftCards = objListToDict("barcode", [
    new GiftCard({ barcode: "123", sum: 1000 }),
    new GiftCard({ barcode: "1234", sum: 100000 }),
])

/*
 * Логика работы с чеком
 */

// Элемент чека
class Item {
    constructor(type) {
        this.id = 0
        this.itemType = type
        this.name = ""
    }
}

// Товарная позиция чека
class Position extends Item {
    constructor(product, props) {
        super("POSITION")
        this.product = product
        this.quantity = 1
        this.returnQuantity = 0
        this.weightFromScale = product.requestWeight
        this.posVerified = false
        Object.assign(this, props)
    }

    // Сумма позиции без учета скидки
    sumWithoutDisc() {
        const qty = this.product.fractional ? this.quantity / 1000 : this.quantity
        return Math.ceil(this.product.price * qty)
    }

    // Скидка
    discount() {
        return Math.ceil(this.sumWithoutDisc() * this.product.discount / 100)
    }

    // сумма с учетом скидки
    sum() {
        return this.sumWithoutDisc() - this.discount()
    }

    // Необходима проверка товара
    posVerify() {
        return this.product.posVerify && !this.posVerified
    }

    // Установка флага выполнения проверки
    setPosVerified() {
        this.posVerified = true
    }

    // преобразование в diff для обновления
    toDiff(type) {
        return {
            diff: type,
            id: this.id,
            name: this.product.name,
            barcode: this.product.barcode,
            price: this.product.price,
            quantity: this.product.fractional ? 0 : this.quantity,
            weight: this.product.fractional ? this.quantity : 0,
            unit: this.product.unit,
            sum: this.sum(),
            discount: this.discount(),
            visualVerify: this.product.visualVerify,
            lowWeight: this.product.lowWeight,
            bypassWeightControl: this.product.bypassWeightControl,
            age: this.product.age,
            weightFromScale: this.weightFromScale,
            returnQuantity: this.product.fractional ? 0 : this.returnQuantity,
            returnWeight: this.product.fractional ? this.returnQuantity : 0,
            tags: this.product.tags,
        }
    }
}

// Карта
class Card extends Item {
    constructor(props) {
        super("CARD")
        this.hasBalance = false
        this.balance = 0
        this.earn = 0
        this.canSpend = 0
        Object.assign(this, props)
    }


    // преобразование в diff для обновления
    toDiff(type) {
        return {
            diff: type,
            id: this.id,
            name: this.name,
            hasBalance: this.hasBalnce,
            balance: this.balance,
            earn: this.earn,
        }
    }
}

// Купон
class Coupon extends Item {
    constructor(props) {
        super("COUPON")
        Object.assign(this, props)
    }
}

// Оплата
class Payment extends Item {
    constructor(props) {
        super("PAYMENT")
        this.type = { type: "UNDEFINED" }
        Object.assign(this, props)
        this.sum = Number(this.sum)
    }
}

// Скидка
class Discount extends Item {
    constructor(props) {
        super("DISCOUNT")
        Object.assign(this, props)
    }
}

// Чек
class Check {
    constructor() {
        this.positions = {}     // позиции
        this.posId = 0          // следующий id позиции
        this.cards = {}         // карты
        this.cardId = 0         // следующий id карты
        this.coupons = {}       // примененные купоны
        this.couponId = 0       // следующий id примененного купона
        this.couponsGiven = {}  // купоны на печать
        this.couponsGivenId = 0 // следующий id купона на печать
        this.payments = {}      // оплаты
        this.paymentId = 0      // следующий id оплаты
        this.discounts = {}     // скидки
        this.discountId = 0     // следующий id скидки
    }

    // Добавление карты в чек
    addCard(props, update = true) {
        const newCard = new Card(props)
        newCard.id = ++this.cardId
        if (newCard.name.includes("999")) {
            newCard.hasBalnce = true
            newCard.balance = 2000
            newCard.earn = 245
        }
        this.cards[newCard.id] = newCard
        if (update)
            this.update({ cards: newCard.toDiff("ADD") })
    }

    // Добавление купона в чек
    addCoupon(props) {
        const newCoupon = new Coupon(props)
        newCoupon.id = ++this.couponId
        this.coupons[newCoupon.id] = newCoupon
        const diff = {
            coupons: {
                id: newCoupon.id,
                diff: "ADD",
                name: newCoupon.name,
            }
        }
        this.update(diff)
    }

    // Добавление купона на печать покупателю
    addCouponGiven(props) {
        const newCoupon = new Coupon(props)
        newCoupon.id = ++this.couponsGivenId
        this.couponsGiven[newCoupon.id] = newCoupon
        const diff = {
            couponsGiven: {
                id: newCoupon.id,
                diff: "ADD",
                name: newCoupon.name,
            }
        }
        this.update(diff)
    }

    // Добавление скидки в чек
    addDiscount(props) {
        const newDisc = new Discount(props)
        newDisc.id = ++this.discountId
        this.discounts[newDisc.id] = newDisc
        const diff = {
            discounts: { id: newDisc.id, diff: "ADD", name: newDisc.name, verify: newDisc.verify }
        }
        this.update(diff)
    }

    // Добавление товара в чек
    addPosition(product, props, update = true) {
        const newPos = new Position(product, props)
        newPos.id = ++this.posId
        if (product.unity) {
            for (const id in this.positions) {
                if (this.positions[id].product.barcode === product.barcode) {
                    newPos.quantity += this.positions[id].quantity
                    delete this.positions[id]
                }
            }
            this.positions[newPos.id] = newPos
            if (update)
                this.update({}, true)
        } else {
            this.positions[newPos.id] = newPos
            if (update)
                this.update({ positions: newPos.toDiff("ADD") })
        }
    }

    // Изменение количества позиции
    changeQuantity(id, quantity) {
        const pos = check.positions[id]
        pos.quantity = quantity
        const diff = { positions: pos.toDiff("UPDATE") }
        this.update(diff)
    }

    // Добавление оплаты
    addPayment(props) {
        if (props.type.type === "CASH") {
            for (const id in this.payments) {
                const payment = this.payments[id]
                if (payment.type.type !== "CASH")
                    continue

                payment.sum += Number(props.sum)
                const diff = {
                    payments: {
                        id: payment.id,
                        diff: "UPDATE",
                        name: payment.name,
                        sum: payment.sum,
                        type: payment.type,
                    }
                }
                this.update(diff)
                return
            }
        }

        const newPayment = new Payment(props)
        newPayment.id = ++this.paymentId
        this.payments[newPayment.id] = newPayment
        const diff = {
            payments: {
                id: newPayment.id,
                diff: "ADD",
                name: newPayment.name,
                sum: newPayment.sum,
                type: newPayment.type,
            }
        }
        this.update(diff)
    }

    // Сумма чека с учетом скидок
    total() {
        let sum = 0
        for (const id in this.positions)
            sum += this.positions[id].sum()
        return sum
    }

    // Сумма скидок
    discount() {
        let disc = 0
        for (const id in this.positions)
            disc += this.positions[id].discount()
        return disc
    }

    // Оплачено
    paid() {
        let sum = 0
        for (const id in this.payments)
            sum += this.payments[id].sum
        return sum
    }

    // К оплате
    due() {
        return Math.max(0, this.total() - this.paid())
    }

    // Необходимы проверки товаров
    posVerify() {
        for (const id in this.positions) {
            if (this.positions[id].posVerify())
                return true
        }
        return false
    }

    // Обновление чека, отправка изменений на SCO
    update(diff, replace = false) {
        diff.diff = !replace
        diff.total = this.total()
        diff.discount = this.discount()
        diff.paid = this.paid()
        diff.paymentTypes = this.paymentTypes()
        diff.bonusToSpend = Math.ceil(diff.total / 2)
        diff.checkItems = this.posVerify()
        if (replace) {
            const diffItems = ["positions", "cards", "coupons", "couponsGiven", "payments", "discounts"]
                .reduce((obj, name) => {
                    const items = this[name]
                    obj[name] = []
                    for (const id in items)
                        obj[name].push(items[id].toDiff("ADD"))
                    return obj
                }, {})
            Object.assign(diff, diffItems)
        }
        call("updateCheck", diff)
    }

    // Доступные типы оплаты
    paymentTypes() {
        const types = [{ type: "CASH" }, { type: "BANKCARD" }, { type: "QR" }, { type: "QRSCAN" }]
        if (Object.keys(this.cards).length > 0 && !this.hasPayment("BONUS"))
            types.push({ type: "BONUS" })
        if (!this.hasPayment("GIFTCARD"))
            types.push({ type: "GIFTCARD" })
        return types
    }

    // Проверка наличия позиции со ШК
    hasPosition(barcode) {
        for (const id in this.positions) {
            const pos = this.positions[id]
            if (pos.product.barcode === barcode)
                return true
        }
        return false
    }

    // Поиск позиции со ШК
    findPosition(barcode) {
        for (const id in this.positions) {
            const pos = this.positions[id]
            if (pos.product.barcode === barcode)
                return pos
        }
        return null
    }

    // Проверка наличия позиции со ШК
    hasPayment(type) {
        for (const id in this.payments) {
            const payment = this.payments[id]
            if (payment.type.type === type)
                return true
        }
        return false
    }

    // Удаление элемента чека
    removeItem(type, id) {
        delete this.itemsByType(type)[id]
        const diff = { [type.toLowerCase() + "s"]: { id, diff: "REMOVE" } }
        this.update(diff)
    }

    // Элементы чека по типу
    itemsByType(type) {
        return this[type.toLowerCase() + "s"]
    }

    // Преобразование чека выбора количества для возврата в чек возврата
    toReturn() {
        for (const id in this.positions) {
            const pos = this.positions[id]
            if (pos.returnQuantity <= 0) {
                delete this.positions[id]
            } else {
                pos.quantity = pos.returnQuantity
                pos.returnQuantity = 0
            }
        }
        this.update({}, true)
    }
}

// Текущий чек
let check = new Check()


/*
 * Обработка событий
 */

let callId = 1
const getCallId = () => callId++

function onEvFailure() {
    check = new Check()
}

function onEvInit() {
    call("initComplete")
}

function onEvAuth(event) {
    const input = String(event.input)
    const params = { result: false }
    if (input in users) {
        user = users[input]
        Object.assign(params, { result: true, cashierName: user.name })
    }
    call("authResult", params)
}

function onEvAddItem(event) {
    let input = String(event.input)
    const result = { forCustomer: true, result: "OK" }
    const props = {}
    if (input.startsWith("22") && input.length >= 7) {
        if (input.length == 7) {
            result.result = "ERROR"
            result.error = "Ошибка при разборе штрих-кода"
            return call("addItemResult", result)
        } else {
            props.quantity = parseInt(input.substr(2, 5), 10)
            input = input.substr(7)
        }
    }
    try {
        if (event.source == "MSR" || /^;.+|^9.+/.test(input) || event.itemType.startsWith("CARD")) {
            // Карты если с ридера, начинаются с ';' или '9'
            check.addCard({ name: input })
        } else if (/^8.+/.test(input) || event.itemType.startsWith("COUPON")) {
            // Купоны если начинаются с '8'
            check.addCoupon({ name: input })
        } else if (/^7.+/.test(input)) {
            // Скидка если начинаются с '7', c '70' требуется проверка
            check.addDiscount({ name: input, verify: /^70.*/.test(input) })
        } else if (input in products) {
            // Остальное позиции
            const product = products[input]
            props.quantity = Number(event.quantity)
            if (props.quantity === 0)
                delete props.quantity
            if (product.requestQuantity && !props.quantity) {
                const res = call("getItemQuantity", { name: product.name, barcode: input })
                if (res.result) {
                    props.quantity = res.quantity
                } else {
                    result.result = "CANCELED"
                }
            }
            if (product.requestWeight) {
                const res = call("getItemWeight", { name: product.name })
                if (res.result) {
                    props.quantity = res.quantity
                } else {
                    result.result = "CANCELED"
                }
            }
            if (product.egais) {
                const res = call("getItemMark", { name: product.name, type: "EGAIS" })
                if (!res.result)
                    result.result = "CANCELED"
            }
            if (product.mark) {
                const res = call("getItemMark", { name: product.name, type: "DATAMATRIX" })
                if (!res.result)
                    result.result = "CANCELED"
            }
            if (result.result === "OK") {
                if (product.barcode === "26")
                    check.addCouponGiven({ name: "Скидка 100500% на все" })
                check.addPosition(product, props)
            }
        } else {
            result.result = "ERROR"
            result.error = "addItemProductNotFound"
        }
        call("addItemResult", result)
    } catch (e) {
        if (e.name == "Cancelled") {
            call("addItemResult", { result: "CANCELED" })
        } else {
            throw e
        }
    }
}

function onEvSubtotal() {
    const params = { result: "OK" }
    if (check.hasPosition("7")) {
        params.result = "ERROR"
        params.forCustomer = true
        const posName = products["7"].name
        params.error = `Добавлена позиция '${posName}'`
    }
    call("subtotalResult", params)
}

function onEvBackToForming() {
    const params = { result: "OK" }
    if (check.hasPosition("8")) {
        params.result = "ERROR"
        params.forCustomer = true
        const posName = products["8"].name
        params.error = `Добавлена позиция '${posName}'`
    }
    call("backToFormingResult", params)
}

function onEvAddPayment(event) {
    const result = { forCustomer: true, result: "OK" }
    let timeout = 0

    try {
        switch (event.type.type) {
        case "CASH":
            check.addPayment({ name: "наличные", type: event.type, sum: event.sum })
            break
        case "BONUS":
            check.addPayment({ name: "бонусами", type: event.type, sum: event.sum })
            break
        case "BANKCARD":
            timeout = 3000
            if (check.hasPosition("10")) {
                result.result = "ERROR"
                result.error = "Недостаточно средств"
            } else if (check.hasPosition("25")) {
                result.result = "ERROR"
                result.error = "Ошибка терминала"
                result.forCustomer = false
            } else {
                check.addPayment({ name: "банковской картой", type: event.type, sum: check.due() })
            }
            break
        case "QR": {
            const rep = call("getPaymentQr")
            if (!rep.result)
                break
            if (rep.input.includes("666")) {
                result.result = "ERROR"
                result.error = "Ошибка оплаты QR-кодом"
            } else {
                check.addPayment({ name: "qr-кодом", type: event.type, sum: check.due() })
            }
            break
        }
        case "QRSCAN": {
            let rep = null
            const callId = getCallId()
            setTimeout(15000, () => {
                if (rep) return
                call("cancelCall", { callId })
            })

            try {
                rep = call("showPaymentQr", { callId, timeout: 15, showCancel: true, data: "some random qr code data 1234567890 1234567890 1234567890" })
                if (rep.userCancel)
                    result.result = "CANCELED"
            } catch (e) {
                if (e.name === "Cancelled") {
                    check.addPayment({ name: "по qr-коду", type: event.type, sum: check.due() })
                } else {
                    throw e
                }
            }
            break
        }
        case "GIFTCARD": {
            if (!(event.input in giftCards)) {
                result.result = "ERROR"
                result.error = "giftCardNotFound"
                break
            }
            const giftCard = giftCards[event.input]
            const sum = Math.min(giftCard.sum, check.due())
            check.addPayment({ name: "подарочной картой", type: event.type, sum })
            break
        }
        }
        setTimeout(timeout, () => {
            call("addPaymentResult", result)
        })
    } catch (e) {
        if (e.name == "Cancelled") {
            call("addPaymentResult", { result: "CANCELED" })
        } else {
            throw e
        }
    }
}

function onEvGetProduct(event) {
    const result = {
        result: true,
        barcode: event.input,
        price: 0,
        quantity: 0,
        weight: 0,
        requestWeight: false,
        lowWeight: false,
        bypassWeightControl: false,
        unit: "",
    }

    if (result.barcode.startsWith("21") && result.barcode.length >= 5) {
        if (result.barcode.length == 5) {
            result.result = false
            result.error = "Ошибка при разборе штрих-кода"
        } else {
            result.quantity = parseInt(result.barcode.substr(2, 3), 10)
            result.barcode = result.barcode.substr(5)
        }
    } else if (result.barcode.startsWith("22") && result.barcode.length >= 7) {
        if (result.barcode.length == 7) {
            result.result = false
            result.error = "Ошибка при разборе штрих-кода"
        } else {
            result.weight = parseInt(result.barcode.substr(2, 5), 10)
            result.barcode = result.barcode.substr(7)
        }
    }

    if (result.result) {
        if (result.barcode in products) {
            const product = products[result.barcode]
            result.name = product.name
            result.price = product.price
            result.requestWeight = product.requestWeight
            result.lowWeight = product.lowWeight
            result.bypassWeightControl = product.bypassWeightControl
            result.unit = product.unit
        } else {
            result.result = false
            result.error = "addItemProductNotFound"
        }
    }
    call("getProductResult", result)
}

function onEvClose() {
    const params = { result: "OK" }
    if (check.hasPosition("11")) {
        params.result = "ERROR"
        const posName = products["11"].name
        params.error = `Добавлена позиция '${posName}'`
    } else {
        if (check.hasPosition("24"))
            params.receipt = "EMAIL"
        check = new Check()
    }

    setTimeout(100, () => call("setProgress", { value: 20 }))
    setTimeout(200, () => call("setProgress", { value: 40 }))
    setTimeout(300, () => call("setProgress", { value: 60 }))
    setTimeout(400, () => call("setProgress", { value: 80 }))
    setTimeout(500, () => call("closeResult", params))
}

function onRemoveItem(event) {
    const params = { result: "OK" }
    if (event.type == "POSITION" && check.positions[event.id].product.barcode == "12") {
        params.result = "ERROR"
        const posName = products["12"].name
        params.error = `Удаление позиции '${posName}'`
    } else {
        check.removeItem(event.type, event.id)
    }
    call("removeItemResult", params)
}

function onChangeQuantity(event) {
    const params = { result: "OK" }
    const pos = check.positions[event.id]
    const product = pos.product

    try {
        if (product.barcode === "14") {
            params.result = "ERROR"
            params.error = "Изменение количества запрещено"
        } else {
            if (event.quantity > 0) {
                check.changeQuantity(event.id, event.quantity)
            } else {
                const result = call("getItemQuantity", { 
                    name: product.name,
                    fractional: product.fractional,
                    quantity: pos.quantity,
                })
                if (!result.result) {
                    params.result = "CANCELED"
                } else {
                    let quantity = Number(result.quantity)
                    if (quantity <= 0) {
                        params.result = "ERROR"
                        params.error = "Количество не может быть меньше 1"
                    } else {
                        check.changeQuantity(event.id, quantity)
                    }
                }
            }
        }
        call("changeQuantityResult", params)
    } catch (e) {
        if (e.name == "Cancelled") {
            call("changeQuantityResult", { result: "CANCELED" })
        } else {
            throw e
        }
    }
}

function onCancelCheck() {
    const params = { result: "OK" }
    if (check.hasPosition("13")) {
        params.result = "ERROR"
        const posName = products["13"].name
        params.error = `Отмена чека с позицией '${posName}'`
    } else {
        check = new Check()
    }
    call("cancelCheckResult", params)
}

const shifts = []
for (let i = 0; i < 256; i++) {
    let number = "" + (i + 1)
    shifts.push({ id: "" + i, cells: [number, "01.01.1970", "01.01.1970", Object.values(users)[i % 3].name] })
}

function addRowsTable(request, from = 0, filter = "") {
    const count = 40
    let items = shifts.filter(item => ~item.cells[0].indexOf(filter))
    let i = from
    for (; i < items.length && request.rows.length < count; i++)
        request.rows.push(items[i])
    request.more = i < items.length
}

function showTable(message) {
    const request = {
        showFilter: true,
        title: message,
        text: "Выберите смену",
        rows: [],
        headers: [
            { name: "Номер", width: -2, align: "CENTER" },
            { name: "Открыта", width: -2, align: "CENTER" },
            { name: "Закрыта", width: -2, align: "CENTER" },
            { name: "Кассир", width: -1, align: "LEFT" },
        ],
        showHeaders: true,
        more: true,
    }
    addRowsTable(request)
    return call("table", request)
}

function onUpdateTable(event) {
    const request = {
        rows: [],
        more: true,
    }
    addRowsTable(request, event.from, event.filter)
    call("updateTable", request)
}

const reports = { X: "X отчет", XFR: "X отчет (ФР)", DEPT: "Отчет по сторно", EMV: "Сверка итогов по безналу" }

function onAction(event) {
    var args = event.args.reduce((obj, cur) => {
        obj[cur.name] = cur.value
        return obj
    }, {})
    const command = args.command

    switch (command) {
    case "SHIFTCLOSE":
        call("setProgress", { value: 30, text: "Выгрузка информации о продажах" })
        setTimeout(3000, () => {
            call("setProgress", { value: 70, text: "Закрытие смены в ФР" })
            setTimeout(2000, () => {
                call("actionResult", { result: "OK" })
            })
        })
        break
    case "REPORT": {
        if (!(args.type in reports)) {
            call("actionResult", { result: "ERROR", error: "Неверный тип отчета: " + args.type })
            break
        }
        const message = reports[args.type]
        const reply = showTable(message)
        if (reply.result) {
            call("actionResult", { result: "OK" })
            call("message", { title: message, text: "Выбрана смена: " + reply.id })
            break
        }
        call("actionResult", { result: "ERROR", error: "Ничего не выбрано" })
        break
    }
    case "MANUAL_DISCOUNT":
        try {
            if (call("choice", { forCustomer: true, title: args.title, text: args.message }).result)
                check.addDiscount({ name: args.name, verify: true })
            call("actionResult", { result: "OK" })
        } catch (e) {
            if (e.name == "Cancelled") {
                call("actionResult", { result: "CANCELED" })
            } else {
                throw e
            }
        }
        break
    default:
        call("actionResult", { result: "ERROR", error: "Неизвестная команда: " + command })
        break
    }
}

function onAccessTest({ command }) {
    switch (user.id) {
    case "1":
        if (["CHECK_EDIT", "CHECK_CANCEL", "CHECK_REMOVEITEM"].includes(command)) {
            call("accessTestResult", { result: "FORBIDDEN" })
            return
        }
        break
    case "2":
        if (command === "MENU_CLOSE") {
            call("accessTestResult", { result: "FORBIDDEN" })
            return
        }
        break
    default:
        break
    }
    call("accessTestResult", { result: "OK" })
}

// eslint-disable-next-line no-unused-vars
function onShutdown(event) {
    setTimeout(2000, () => {
        call("shutdownComplete")
    })
}

// Получить абсолютный путь до изображения из каталога
function imgPath(img) {
    // eslint-disable-next-line no-undef
    return { type: "URL", data: appDirPath + "/catalog/" + img }
}

// Элемент каталога товаров
class CatalogItem {
    constructor(props) {
        this.name = ""          // наименование
        this.image = ""         // путь до изображения или изображение в формате base64
        this.isGroup = false    // является ли данный элемент группой товаров
        this.id = ""            // id группы или штрих-код товара
        this.price = 0          // цена
        this.unit = ""          // единицы измерения
        Object.assign(this, props)
    }
}

const catalog = [
    new CatalogItem({ name: "Тестовые товары", image: imgPath("c_1.png"), isGroup: true, id: "1" }),
    new CatalogItem({ name: "Спец. товары", image: imgPath("c_2.png"), isGroup: true, id: "2" }),
]

function makeItem(barcode) {
    const product = products[barcode]
    return new CatalogItem({ name: product.name, image: imgPath(barcode + ".png"), id: barcode, price: product.price, unit: product.unit })
}

const catalogGoods = {
    "1": [makeItem("1"), makeItem("2"), makeItem("3")],
    "2": [makeItem("4"), makeItem("5"), makeItem("7"), makeItem("8"), makeItem("9"), makeItem("10"), makeItem("11"),
        makeItem("15"), makeItem("16"), makeItem("19"), makeItem("20")],
}

function onGetPickList(event) {
    let items = []
    if (event.codes && event.codes.length != 0) {
        for (let group in catalogGoods) {
            items.push(catalogGoods[group].filter(item => (event.codes.includes(item.id) && ~item.name.indexOf(event.filter))))
        }
    } else if (event.groupId && !event.filter) {
        items = catalogGoods[event.groupId]
    } else if (event.groupId) {
        items = catalogGoods[event.groupId].filter(item => ~item.name.indexOf(event.filter))
    } else if (event.filter) {
        items.push(catalog.filter(item => ~item.name.indexOf(event.filter)))
        for (let group in catalogGoods) {
            items.push(catalogGoods[group].filter(item => ~item.name.indexOf(event.filter)))
        }
    } else {
        items = catalog
    }

    call("getPickListResult", { items: items })
}

function onSetCustomerAddress(event) {
    const params = { result: "OK" }
    if (event.address == "+79999999999") {
        params.result = "ERROR"
        params.error = "Неверный номер телефона"
        params.forCustomer = true
    }
    call("setCustomerAddressResult", params)
}

let cashBalance = 1000000
function onGetCashBalance() {
    call("getCashBalanceResult", { sum: cashBalance })
}

function onMoneyOperation(event) {
    const sum = Number(event.sum)
    cashBalance = cashBalance + (event.out ? -sum : sum)
    call("moneyOperationResult", { result: "OK" })
}

const menu = {
    "1": [
        { text: "Редактирование чека", command: "CHECK_EDIT" },
        { text: "Тест оборудования", command: "MENU_SUBMENU", args: [{ name: "id", value: "5" }] },
        { text: "Режим кассира", command: "SCO_CASHIERMODE" },
        { text: "Закрытие смены", command: "CORE_CLIENTACTION", args: [{ name: "command", value: "SHIFTCLOSE" }] },
        { text: "Завершение работы", command: "CORE_SHUTDOWN" },
    ],
    "2": [
        { text: "Редактирование чека", command: "CHECK_EDIT" },
        { text: "Режим кассира", command: "SCO_CASHIERMODE" },
    ],
    "5": [
        { text: "Устройства ввода", command: "INPUT_INPUTDEVICESTEST" },
        { text: "Прикассовые весы", command: "INPUT_SCALETEST" },
        { text: "Лампы оповещения", command: "SCO_LIGHTSTEST" },
        { text: "Контрольные весы", command: "WEIGHTCONTROL_SCALETEST" },
    ],
}

function onGetMenu(event) {
    const id = String(event.id)
    const result = {}
    if (id in menu)
        result.items = menu[id]
    call("getMenuResult", result)
}

function onEvPrint() {
    call("setProgress", { value: 30, text: "Печать чека" })
    setTimeout(3000, () => {
        call("printResult", { result: "OK" })
    })
}

function onTrainingMode() {
    call("trainingModeResult", { result: "OK" })
}

function onCheckItems() {
    try {
        while (check.posVerify()) {
            if (call("choice", {
                title: "Проверка товаров", text: "Необходимо подтверждение продажи", okButton: "Подтвердить",
                cancelButton: "Редакт. чек"
            }).result) {
                for (const id in check.positions) {
                    if (check.positions[id].posVerify())
                        check.positions[id].setPosVerified()
                }
                check.update({})
            } else {
                call("editCheck")
            }
        }
        call("checkItemsComplete")
    } catch (e) {
        if (e.name == "Cancelled") {
            call("checkItemsComplete")
        } else {
            throw e
        }
    }
}

function onReturn() {
    const params = { result: "OK" }
    if (call("choice", {
        title: "Выбор чека", text: "Клиент в этот момент может отображать диалоги выбора документа. \n" +
            "Выполнить возврат документа?", okButton: "Ok", cancelButton: "Отмена"
    }).result) {
        check = new Check()
        check.addCard({ name: "999" }, false)
        check.addPosition(products["1"], { quantity: 2 }, false)
        check.addPosition(products["2"], { quantity: 10 }, false)
        check.addPosition(products["3"], { quantity: 1 }, false)
        check.addPosition(products["6"], { quantity: 3293 }, false)
        check.update({}, true)
        params.selectItems = true
    } else {
        params.result = "CANCELED"
    }
    call("returnResult", params)
}

function onReturnSelect(event) {
    const params = { result: "OK" }
    switch (event.type) {
    case "ALL":
        for (const id in check.positions)
            check.positions[id].returnQuantity = check.positions[id].quantity
        check.update({}, true)
        break
    case "RESET":
        for (const id in check.positions)
            check.positions[id].returnQuantity = 0
        break
    case "QUANTITY": {
        const pos = check.positions[event.id]
        const product = pos.product
        const result = call("getItemQuantity", { name: product.name, fractional: product.fractional })
        if (result.result)
            pos.returnQuantity = Math.min(pos.quantity, Math.max(0, result.quantity))
        break
    }
    case "CLEAR": {
        check.positions[event.id].returnQuantity = 0
        break
    }
    case "INPUT": {
        const pos = check.findPosition(event.input)
        if (!pos) {
            params.result = "ERROR"
            params.error = "addItemProductNotFound"
            break
        }
        pos.returnQuantity = Math.min(pos.quantity, Math.max(0, pos.returnQuantity + 1))
        debug("", { rqty: pos.returnQuantity })
        break
    }
    default:
        params.result = "CANCELED"
        break
    }
    check.update({}, true)
    call("returnSelectResult", params)
}

function onReturnSelected() {
    check.toReturn()
    call("returnSelectedResult", { result: "OK" })
}

// Обработчики событий
const eventHandlers = {
    EvInit: onEvInit,
    EvFailure: onEvFailure,
    EvAuth: onEvAuth,
    EvAddItem: onEvAddItem,
    EvSubtotal: onEvSubtotal,
    EvBackToForming: onEvBackToForming,
    EvAddPayment: onEvAddPayment,
    EvGetProduct: onEvGetProduct,
    EvClose: onEvClose,
    EvRemoveItem: onRemoveItem,
    EvChangeQuantity: onChangeQuantity,
    EvCancelCheck: onCancelCheck,
    EvAction: onAction,
    EvAccessTest: onAccessTest,
    EvShutdown: onShutdown,
    EvGetPickList: onGetPickList,
    EvSetCustomerAddress: onSetCustomerAddress,
    EvUpdateTable: onUpdateTable,
    EvCustomerLang: () => { },
    EvGetCashBalance: onGetCashBalance,
    EvMoneyOperation: onMoneyOperation,
    EvGetMenu: onGetMenu,
    EvPrint: onEvPrint,
    EvTrainingMode: onTrainingMode,
    EvCheckItems: onCheckItems,
    EvReturn: onReturn,
    EvReturnSelect: onReturnSelect,
    EvReturnSelected: onReturnSelected,
    EvMode: () => { },
}

// Вызов метода API на SCO
function call(name, params) {
    // eslint-disable-next-line no-undef
    const result = grpc.call(name, params)
    if (result["status"] == 1)
        throw new Cancelled()
    if (result["status"] != 0)
        throw new Error("Ошибка выполнения метода " + name)
    return result.reply
}

/*
 * Экспортируемые функции
 * Вызываются из sst-client
 */

// установлено подключение
let connected = false

// Обрабатываемые события, параметры метода events()
export function eventsRequest() {
    const result = {
        types: Object.keys(eventHandlers),
        reconnect: connected,
    }
    connected = true
    return result
}

// Обработка событий. Вызывается при получении события от SCO
export function onEvent(event) {
    const body = event.body
    const type = body["@type"].replace("type.googleapis.com/sco.", "")
    if (!(type in eventHandlers)) {
        console.error("Не найден обработчик события " + type)
        return
    }
    eventHandlers[type](body)
}
